/* $Id: ide-std.c,v 1.4 1999/06/11 14:29:45 ralf Exp $
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * IDE routines for typical pc-like standard configurations.
 *
 * Copyright (C) 1998, 1999 by Ralf Baechle
 */
#include <linux/sched.h>
#include <linux/ide.h>
#include <linux/ioport.h>
#include <linux/hdreg.h>
#include <asm/ptrace.h>
#include <asm/hdreg.h>
#include <asm/io.h>

#include <linux/tivoconfig.h>

static void (*tivo_cmd_hook)(ide_drive_t *, unsigned char *);
static unsigned int tivo_size_priv(int dma_chunks, int block_chunks);
static void tivo_init_drvpriv(struct idetivo_request *trq, int reading);
static int tivo_add_dmachunk(struct idetivo_request *trq, unsigned long k_addr,
                             unsigned long u_addr, int size, int reading);
static void tivo_add_blkchunk(struct idetivo_request *trq, int chunks, int sectors);
static void tivo_end_blkchunk(struct idetivo_request *trq, int chunks, int sectors);
static void tivo_set_cmd_hook(void (*handler)(ide_drive_t *, unsigned char *)); 

static struct ll_dma_ops tivo_sgops = {
    size_priv   :   tivo_size_priv,
    init_drvpriv:   tivo_init_drvpriv,
    add_dmachunk:   tivo_add_dmachunk,
    add_blkchunk:   tivo_add_blkchunk,
    end_blkchunk:   tivo_end_blkchunk,
    set_cmd_hook:   tivo_set_cmd_hook,
};

struct tivo_sgent {
    u32 address;
    u32 count;
};

struct desc_chunk {
    unsigned int sglist;   //bus address of scatter list
};

struct tivo_ide_priv {
    struct desc_chunk *first_chunk;
    struct desc_chunk *cur_chunk;
    struct tivo_sgent *first_sgent;
    struct tivo_sgent *cur_sgent;
    int big_flush;
    int last_size;
    u32 bus_addr_end;
};

/*
 These functions are used to preprocess TiVo I/O requests

 Report back how much room needed to allocate along with the request
 for use by the DMA engine, based on number of scatter list entries,
 and number of distinct scatter lists.  Each separate IDE transfer
 (of max MAX_NRSECTORS sectors) requires a separate scatter list.
*/
static unsigned int tivo_size_priv(int dma_chunks, int block_chunks)
{
   unsigned int total_chunk_size = dma_chunks * sizeof(struct tivo_sgent);

   unsigned int size = sizeof(struct tivo_ide_priv) +
               block_chunks * sizeof(struct desc_chunk) +
               (block_chunks - 1) * 8 + // worst case PRD table alignment
               L1_CACHE_BYTES - 4 +  // worst case cache line alignment
               total_chunk_size;
    return size;
}

/*
  put the chunk list right after the priv struct
  put the scatter list after the chunk list (but cache line align it)
*/
static void tivo_init_drvpriv(struct idetivo_request *trq, int reading)
{
    struct tivo_ide_priv *priv = trq->drv_priv;
    unsigned long tmp;

    priv->first_chunk = (struct desc_chunk *)((unsigned long)priv +
                                              sizeof(struct tivo_ide_priv));

    tmp = (unsigned long)(priv->first_chunk + trq->num_chunks);
    priv->first_sgent = (struct tivo_sgent *)L1_CACHE_ALIGN(tmp);

    priv->cur_sgent = priv->first_sgent;
    priv->cur_chunk = priv->first_chunk;
    priv->cur_chunk->sglist = virt_to_phys(priv->cur_sgent);

    // this is a bit of platform specificness, but there's no generic way
    // to do it: if the request totals more than twice the dcache size,
    // mark to do one big flush, instead of flushing each individual chunk
    //
    priv->big_flush = (trq->req->nr_sectors > 16 * 1024 / SECTOR_BYTES);

    priv->bus_addr_end = 1; // an impossible addr
}


/*
    Add a chunk to the DMA scatter list and do any necessary cache flushing
    Both kernel and user virtual addresses are passed in, in case the user
    address is needed for flushing purposes (to avoid virtual aliasing
    problem on CPUs with virually indexed caches).  The user address is
    guaranteed not to page fault and size is guaranteed to be non-zero.
*/
static int tivo_add_dmachunk(struct idetivo_request *trq, unsigned long k_addr,
                             unsigned long u_addr, int size, int reading)
{
    struct tivo_ide_priv *priv = trq->drv_priv;
    struct tivo_sgent *sgent = priv->cur_sgent;
    u32 bus_addr;
    int total_size;

    if ((k_addr | size) & 3) {
        printk(KERN_WARNING "tivo_ide: DMA not word aligned\n");
        return -EINVAL;
    }

    if (!priv->big_flush) {
        if (reading) {
            dma_cache_inv(u_addr, size);
        } else {
            dma_cache_wback_inv(u_addr, size);
        }
    }

    bus_addr = (u32)virt_to_bus((void *)k_addr);
    // Consolidate adjacent chunks
    //
    // Note that total size will never exceed 128K, because we start a
    // new scatter list at least every 128K
    //
    total_size = size;
    u32 cur_len = total_size;


    while (cur_len) {
        u32 xcount, bcount = 0x10000 - (bus_addr & 0xffff);
        
        if (bcount > cur_len)
            bcount = cur_len;

        xcount = bcount & 0xffff;

#ifdef CONFIG_TIVO_CPU_TO_BE32
        sgent->address = cpu_to_be32(bus_addr);
        sgent->count = cpu_to_be32(xcount);
#else
        sgent->address = cpu_to_le32(bus_addr);
        sgent->count = cpu_to_le32(xcount);
#endif /* CONFIG_TIVO_CPU_TO_BE32 */
        bus_addr += bcount;
        cur_len -= bcount;

        priv->cur_sgent = sgent + 1;
        sgent = priv->cur_sgent;
    }
    priv->last_size = total_size;
    priv->bus_addr_end = bus_addr + size;
    return 0;
}
// Finish up a block of DMA chunks that makes up a single IDE transfer,
// and start a new one
//
static void tivo_add_blkchunk(struct idetivo_request *trq, int chunks, int sectors)
{
    struct tivo_ide_priv *priv = trq->drv_priv;
    struct desc_chunk *desc = priv->cur_chunk++;

#ifdef CONFIG_TIVO_CPU_TO_BE32
    (priv->cur_sgent - 1)->count |= cpu_to_be32(0x80000000);
#else
    (priv->cur_sgent - 1)->count |= cpu_to_le32(0x80000000);
#endif
    // Align start of PRD table to 16-bytes
    priv->cur_sgent = (struct tivo_sgent *)
        (((unsigned long)priv->cur_sgent + 0xf) & ~0xf);
    (desc + 1)->sglist = virt_to_phys(priv->cur_sgent);

    priv->bus_addr_end = 1; // an impossible addr
}

// Finish up the last block of DMA chunks, and do any final prep on the
// data going to the DMA engine
//
static void tivo_end_blkchunk(struct idetivo_request *trq, int chunks, int sectors)
{
    struct tivo_ide_priv *priv = trq->drv_priv;
    unsigned long sg_start = (unsigned long)priv->first_sgent;
    unsigned long sg_end = (unsigned long)priv->cur_sgent;

#ifdef CONFIG_TIVO_CPU_TO_BE32
    (priv->cur_sgent - 1)->count |= cpu_to_be32(0x80000000);
#else
    (priv->cur_sgent - 1)->count |= cpu_to_le32(0x80000000);
#endif
    if (!priv->big_flush) {
        dma_cache_wback_inv(sg_start, sg_end - sg_start);
    } else {
        flush_dcache();
    }
    priv->cur_chunk = priv->first_chunk;
}

static void tivo_set_cmd_hook(void (*handler)(ide_drive_t *, unsigned char *)) {
    tivo_cmd_hook = handler;
}

ide_startstop_t bcm2_dma_intr(ide_drive_t *drive)
{
    int i, dma_stat;
    byte stat;
    struct request *rq = HWGROUP(drive)->rq;
    struct idetivo_request *trq = rq->tivo;

    dma_stat = HWIF(drive)->dmaproc(ide_dma_end, drive);
    stat = GET_STAT();            /* get drive status */
    if (OK_STAT(stat,DRIVE_READY,drive->bad_wstat|DRQ_STAT)) {
        if (!dma_stat) {
            plog_diskstop(rq);
            if (trq) {
                unsigned int max_nrsectors = trq->max_nrsectors;

                if (rq->nr_sectors > max_nrsectors) {
                    ide_hwif_t *hwif = HWIF(drive);
                    ide_startstop_t startstop;

                    rq->nr_sectors -= max_nrsectors;
                    rq->sector += max_nrsectors;
                    trq->progress += max_nrsectors * SECTOR_BYTES;
                    add_blkdev_randomness(MAJOR(rq->rq_dev));
                    SELECT_DRIVE(hwif, drive);
                    if (ide_wait_stat(&startstop, drive, drive->ready_stat,
                                      BUSY_STAT|DRQ_STAT, WAIT_READY)) {
                        printk("%s: drive not ready for command\n", drive->name);
                        return startstop;
                    }
                    //trq->scramble.recordOffset++;
                    return DRIVER(drive)->do_request(drive, rq, rq->sector
                                                      + trq->partition_start);
                } else {
                    trq->progress += rq->nr_sectors * SECTOR_BYTES;
                ide_end_request(1, HWGROUP(drive));
                trq->status = IDETIVO_DONE;
                return ide_stopped;
                }
            } else {
                for (i = rq->nr_sectors; i > 0;) {
                    i -= rq->current_nr_sectors;
                    ide_end_request(1, HWGROUP(drive));
                }
            }
            return ide_stopped;
        }
        printk("%s: dma_intr: bad DMA status\n", drive->name);
    }
    if (trq) {
        struct tivo_ide_priv *priv = trq->drv_priv;

        // if the request gets retried, we need to reprocess this
        // chunk of the DMA scatterlist
        priv->cur_chunk--;
        trq->status = IDETIVO_ERROR;
        trq->drive_status = stat;
        // XXX: should probably read drive regs to query how far we
        // got in last 256 sector (max) chunk, too...
    }
    return ide_error(drive, "dma_intr", stat);
}

int bcm2_build_dmatable(ide_drive_t *drive) 
{
    struct request *rq = HWGROUP(drive)->rq;
    
    if (!rq->tivo) {
        return 1;
    }
    struct tivo_ide_priv *priv = rq->tivo->drv_priv;
    struct desc_chunk *chunk = priv->cur_chunk++;
    struct io_scramble *pscramble = &rq->tivo->scramble;
    ide_hwif_t *hwif = HWIF(drive);

    rq->tivo->status = IDETIVO_INPROGRESS;

    unsigned long dma_base        = hwif->dma_base;
    /* write the PRD table to the controller */
    outl((unsigned long)chunk->sglist, dma_base + 4); 
    barrier();

    return 1;
}

#ifdef BCM_IDE_DEBUG
static inline void print_sg_list(struct tivo_ide_priv *priv)
{
    struct tivo_sgent *sgent = priv->first_sgent;
    unsigned long sg_end = (unsigned long)priv->cur_sgent;

    while ((unsigned long)sgent != sg_end) {
        unsigned long sg_address = be32_to_cpu(sgent->address);
        unsigned long sg_count = sgent->count;
        if (sg_count & 0x80000000) {
            sg_count &= 0x7FFFFFFF;
            sg_count = be32_to_cpu(sg_count);
        }
        printk("sg_address = 0x%x sg_count = %d\n", sg_address, sg_count);
        sgent++;
    }
}
#endif /* BCM_IDE_DEBUG */

static int tivo_ide_default_irq(ide_ioreg_t base)
{
	return 0;
}

static ide_ioreg_t tivo_ide_default_io_base(int index)
{
	return 0;
}

static void tivo_ide_init_hwif_ports(hw_regs_t *hw, ide_ioreg_t data_port,
                                    ide_ioreg_t ctrl_port, int *irq)
{
	ide_ioreg_t reg = data_port;
	int i;
	TivoConfigValue ide_cap;

	for (i = IDE_DATA_OFFSET; i <= IDE_STATUS_OFFSET; i++) {
		hw->io_ports[i] = reg;
		reg += 1;
	}
	if (ctrl_port) {
		hw->io_ports[IDE_CONTROL_OFFSET] = ctrl_port;
	} else {
		hw->io_ports[IDE_CONTROL_OFFSET] = hw->io_ports[IDE_DATA_OFFSET] + 0x206;
	}
	if (irq != NULL)
		*irq = 0;
	hw->io_ports[IDE_IRQ_OFFSET] = 0;
#if 1	
	if (GetTivoConfig(kTivoConfigIdeCap, &ide_cap)) {
	    ide_hwifs[0].udma_four = 0;
	} else {
	    ide_cap &= kTivoConfigIdeCapAta66;
	    ide_hwifs[0].udma_four = ide_cap ? 1 : 0;
	}
#endif
	ide_hwifs[0].tivo.version = IDETIVO_DRIVER_VERSION;
	ide_hwifs[0].tivo.sgops = &tivo_sgops;
}

static int tivo_ide_request_irq(unsigned int irq,
                                void (*handler)(int,void *, struct pt_regs *),
                                unsigned long flags, const char *device,
                                void *dev_id)
{
	return request_irq(irq, handler, flags, device, dev_id);
}			

static void tivo_ide_free_irq(unsigned int irq, void *dev_id)
{
	free_irq(irq, dev_id);
}

static int tivo_ide_check_region(ide_ioreg_t from, unsigned int extent)
{
	return check_region(from, extent);
}

static void tivo_ide_request_region(ide_ioreg_t from, unsigned int extent,
                                    const char *name)
{
	request_region(from, extent, name);
}

static void tivo_ide_release_region(ide_ioreg_t from, unsigned int extent)
{
	release_region(from, extent);
}

struct ide_ops bcm_ide_ops = {
	&tivo_ide_default_irq,
	&tivo_ide_default_io_base,
	&tivo_ide_init_hwif_ports,
	&tivo_ide_request_irq,
	&tivo_ide_free_irq,
	&tivo_ide_check_region,
	&tivo_ide_request_region,
	&tivo_ide_release_region
};

void _tivo_ide_out_byte(u8 data, ide_ioreg_t port)
{
  *((volatile unsigned char *)(mips_io_port_base + port)) = __ioswab32(data);
}

u8 _tivo_ide_in_byte(ide_ioreg_t port)
{
  return  __ioswab32(*((volatile unsigned char *)(mips_io_port_base + port)));
}

u16 _tivo_ide_in_word(ide_ioreg_t port)
{
  return  __ioswab32(*((volatile unsigned short *)(mips_io_port_base + port)));
}
     

/*
 * These are duplicates of code elsewhere, we need it here because it's
 * #ifndef CONFIG_TIVO'ed where TiVo platforms need it to be different,
 * but for PX, we actually need the standard PCI IDE code:
 */

#undef  insw
#define insw(port, addr, count) ide_insw(port, addr, count)
static inline void ide_insw(unsigned long port, void *addr, unsigned int count)
{
  while(count--) {
    *(u16 *)addr =  __ioswab32(*(volatile u16 *)(mips_io_port_base + port));
    addr += 2;
  }
}

#undef  outw
#define outw(port, addr, count) ide_outw(port, addr, count)
static inline void ide_outw(unsigned long port, void *addr, unsigned int count)
{
  while(count--) {
    *(volatile u16 *)(mips_io_port_base + port) =  __ioswab32(*(u16 *) addr);
    addr += 2;
  }
}

#undef  insl
#define insl(port, addr, count) ide_insl(port, addr, count)
static inline void ide_insl(unsigned long port, void *addr, unsigned int count)
{
  while(count--) {
    *(u32 *)addr =  __ioswab32(*(volatile u32 *)(mips_io_port_base + port));
    addr += 4;
  }
}

#undef  outl
#define outl(port, addr, count) ide_outl(port, addr, count)
static inline void ide_outl(unsigned long port, void *addr, unsigned int count)
{
  while(count--) {
    *(volatile u32 *)(mips_io_port_base + port) =  __ioswab32(*(u32 *) addr);
    addr += 4;
  }
}


/*
 * This is used for most PIO data transfers *from* the IDE interface
 */
void ide_input_data (ide_drive_t *drive, void *buffer, unsigned int wcount)
{
	byte io_32bit = drive->io_32bit;

	if (io_32bit) {
		insl(IDE_DATA_REG, buffer, wcount);
	} else {
		insw(IDE_DATA_REG, buffer, wcount<<1);
	}
}

/*
 * This is used for most PIO data transfers *to* the IDE interface
 */
void ide_output_data (ide_drive_t *drive, void *buffer, unsigned int wcount)
{
	byte io_32bit = drive->io_32bit;

	if (io_32bit) {
		outsl(IDE_DATA_REG, buffer, wcount);
	} else {
		outsw(IDE_DATA_REG, buffer, wcount<<1);
	}
}

/*
 * The following routines are mainly used by the ATAPI drivers.
 *
 * These routines will round up any request for an odd number of bytes,
 * so if an odd bytecount is specified, be sure that there's at least one
 * extra byte allocated for the buffer.
 */
void atapi_input_bytes (ide_drive_t *drive, void *buffer, unsigned int bytecount)
{
	++bytecount;
	ide_input_data (drive, buffer, bytecount / 4);
	if ((bytecount & 0x03) >= 2)
		insw (IDE_DATA_REG, ((byte *)buffer) + (bytecount & ~0x03), 1);
}

void atapi_output_bytes (ide_drive_t *drive, void *buffer, unsigned int bytecount)
{
	++bytecount;
	ide_output_data (drive, buffer, bytecount / 4);
	if ((bytecount & 0x03) >= 2)
		outsw (IDE_DATA_REG, ((byte *)buffer) + (bytecount & ~0x03), 1);
}
