/*
 * IRQ-handling code for TiVo devices
 *
 * Copyright (C) 1992 Linus Torvalds
 * Copyright (C) 1994 - 2000 Ralf Baechle
 * Copyright (C) 2000-2005 TiVo Inc. All Rights Reserved.
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/kernel_stat.h>
#include <linux/slab.h>
#include <linux/random.h>
#include <linux/plog.h>
#include <linux/module.h>

#include <asm/mipsregs.h>
#include <asm/ptrace.h>
#include <asm/bcm7038.h>
#include <asm/atomic.h>

#define I2C_COMPATABLE_TEMP   //temptemp XXX <-- Shane: clean up after 06/20/05

extern asmlinkage void tivo_handle_irq(void);
 
atomic_t irq_err_count;

static struct irqaction *irq_action[BCM_NR_IRQS];
static int irq_depth[BCM_NR_IRQS];

#ifdef CONFIG_TIVO_DEVEL
int debug_irqs;

static int __init dbgirq_setup(char *str)
{
	get_option(&str, &debug_irqs);
	return 1;
}

__setup("dbgirq=", dbgirq_setup);

#define DEBUG_IRQS debug_irqs

#else
#define DEBUG_IRQS 0
#endif

static inline
int              // = -1 -- input value is zero, when doing (31 - x)
                 // = 63 -- input value is zero, when doing (31 XOR x)
                 //
                 // > 0  -- 31 - most significant left bit is 1
                 //         ...
                 //          0 - most significant right bit is 1
                 //
get_leading_one_bit_number(unsigned long val)
{
    int bit;

    // clz $t0,$t1 count leading zeros, t0 = # of leading zeros in t1
    //
    //    0x80000000 --> count of leading zeros produce 0
    //    0x40000000 --> count of leading zeros produce 1
    //    ...
    //    0x00000001 --> count of leading zeros produce 31
    //    0x00000000 --> count of leading zeros produce 32
    //
    __asm__ __volatile__ ( ".set mips32\n             \
                             clz %0,%1\n              \
                            .set mips2\n"             \
                         : "=r" (bit) : "r" (val) );
//  return 31 - bit;
    return 31 ^ bit;
}

static inline void mask_irq(unsigned int irq)
{
	unsigned int status;
	if ( irq < 8 )
	{
	    status = read_32bit_cp0_register(CP0_STATUS);
	    status &= ~(1 << (irq + 8));
	    write_32bit_cp0_register(CP0_STATUS, status);
	}
	else if ( irq < 40 )
	{
	    *BCM_INTR_WIN0_MASK_SET = (1 << (irq - 8));
	}
	else if ( irq < 72)
	{
	    *BCM_INTR_WIN1_MASK_SET = (1 << (irq - 40));
	}
	else if ( irq < 104 )
	{
	    *BCM_INTR_IRQ0_ENABLE &= ~(1 << (irq - 72));
	}
	else if ( irq <  136 )
	{
	    *BCM_INTR2_CPU_MASK_SET = (1 << (irq - 104));
	}
	barrier();
}

static inline void ack_irq(unsigned int irq)
{
  // currently not needed -- GIO (when supported) may require additional logic
}

static inline void mask_and_ack_irq(unsigned int irq)
{
  if( irq <= BCM_IRQ_SYSTIMER_INTR ) {
      mask_irq(irq);
      ack_irq(irq);
  }
}

static inline void unmask_irq(unsigned int irq)
{
	unsigned long status;
	if ( irq < 8 )
	{
	    status = read_32bit_cp0_register(CP0_STATUS);
	    status |= (1 << (irq + 8) );
	    write_32bit_cp0_register(CP0_STATUS, status);
	}
	else if ( irq < 40 )
	{
	    *BCM_INTR_WIN0_MASK_CLEAR = (1 << (irq - 8));
	}
	else if ( irq < 72)
	{
	    *BCM_INTR_WIN1_MASK_CLEAR = (1 << (irq - 40));
	}
	else if ( irq < 104 )
	{
	    *BCM_INTR_IRQ0_ENABLE |= (1 << (irq - 72));
	}
	else if ( irq <  136 )
	{
	    *BCM_INTR2_CPU_MASK_CLEAR = (1 << (irq - 104));
	}
	barrier();
}

void disable_irq(unsigned int irq_nr)
{
	unsigned long flags;

	if (DEBUG_IRQS >= 5) {
		printk("dis<%d>", irq_nr);
	}
	if (irq_nr >= BCM_NR_IRQS)
		return;

	save_and_cli(flags);
	if (irq_depth[irq_nr]++ < 0) {
       mask_irq(irq_nr);
    }
	restore_flags(flags);
}

void enable_irq(unsigned int irq_nr)
{
	unsigned long flags;

	if (DEBUG_IRQS >= 5) {
		printk("ena<%d>", irq_nr);
	}
	if (irq_nr >= BCM_NR_IRQS)
		return;
	save_and_cli(flags);
	if (--irq_depth[irq_nr] < 0)
		unmask_irq(irq_nr);
	restore_flags(flags);
}

EXPORT_SYMBOL(disable_irq);
EXPORT_SYMBOL(enable_irq);

int get_irq_list(char *buf)
{
	int i, len = 0;
	struct irqaction *action;

	for (i = 0 ; i < BCM_NR_IRQS ; i++) {
		action = irq_action[i];
		if (!action)
			continue;
		len += sprintf(buf + len, "%2d: %8d %c %s",
			i, kstat.irqs[0][i],
			(action->flags & SA_INTERRUPT) ? '+' : ' ',
			action->name);
		for (action = action->next; action; action = action->next) {
			len += sprintf(buf + len, ",%s %s",
				(action->flags & SA_INTERRUPT) ? " +" : "",
				action->name);
		}
		len += sprintf(buf + len, "\n");
	}
	return len;
}

/*
 * do_IRQ is called by the low-level interrupt handling code with the
 * appropriate IRQ number
 */

asmlinkage void do_IRQ(int irq, struct pt_regs *regs)
{
	struct irqaction *action;
	int cpu;

	mask_and_ack_irq(irq);
	cpu = smp_processor_id();
	irq_enter(cpu, irq);
	kstat.irqs[cpu][irq]++;

	action = irq_action[irq];

	if (DEBUG_IRQS >= 2) {
		if (irq == BCM_IRQ_UPG_TMR_CPU_INTR && DEBUG_IRQS < 5) {
			static unsigned int count1;
			if (!(count1++ & 0xff)) {
				printk("*");
			}
#if 0 // add later for often-firing brcm int
		} else if (irq == BCM_LINUX_BGE_IRQ && DEBUG_IRQS < 5) {
			static unsigned int count2;
			if (!(count2++ & 0xff)) {
				printk("x");
			}
#endif
		} else if (irq == BCM_IRQ_USB_CPU_INTR && DEBUG_IRQS < 5) {
			static unsigned int count3;
			if (!(count3++ & 0xff)) {
				printk("+");
			}
		} else {
			printk("[%d", irq);
		}
	}

	if (action != 0) {
		unsigned long flags = 0;

		plog_intr_enter(irq, regs->cp0_epc);
		do {
			if (!(action->flags & SA_INTERRUPT))
				__sti();
			action->handler(irq, action->dev_id, regs);
			__cli();
			flags |= action->flags;
			action = action->next;
		} while (action);
		if (flags & SA_SAMPLE_RANDOM)
			add_interrupt_randomness(irq);
		// if IRQ was disabled from within its own handler(s),
		// leave it masked, otherwise re-enable
		if (irq_depth[irq] < 0)
			unmask_irq(irq);
		plog_intr_leave(irq);
	} else if (DEBUG_IRQS >= 2) {	// report as unhandled
		printk("!");
	}
	if (DEBUG_IRQS >= 2) {
		if ((irq != BCM_IRQ_UPG_TMR_CPU_INTR &&
#if 0 // add later for often-firing brcm int
		     irq != BCM_LINUX_BGE_IRQ &&
#endif
		     irq != BCM_IRQ_USB_CPU_INTR) ||
		    DEBUG_IRQS >= 5) {
			printk("]");
		}
	}


	irq_exit(cpu, irq);

	if (softirq_pending(cpu))
		do_softirq();
}

int tivo_setup_irq(int irq, struct irqaction *new)
{
	int shared = 0;
	struct irqaction *old, **p;
	unsigned long flags;

	p = irq_action + irq;
	if ((old = *p) != NULL) {
		// Shared ints must all agree to share
		if (!(old->flags & new->flags & SA_SHIRQ))
			return -EBUSY;
		do {
			p = &old->next;
			old = *p;
		} while (old);
		shared = 1;
	}
	if (new->flags & SA_SAMPLE_RANDOM)
		rand_initialize_irq(irq);

	save_and_cli(flags);
	*p = new;

	if (!shared) {
		irq_depth[irq] = -1;
		ack_irq(irq);
		unmask_irq(irq);
	}
	restore_flags(flags);
	return 0;
}

int request_irq(unsigned int irq,
	void (*handler)(int, void *, struct pt_regs *),
	unsigned long irqflags,
	const char *devname,
	void *dev_id )
{
	int retval;
	struct irqaction *action;

	if (DEBUG_IRQS >= 1) {
		printk("request irq %d\n", irq);
	}
	if (irq >= BCM_NR_IRQS)
		return -EINVAL;
	if (!handler)
		return -EINVAL;

	action = (struct irqaction *) kmalloc(sizeof(struct irqaction), GFP_KERNEL);
	if (!action)
		return -ENOMEM;

	action->handler = handler;
	action->flags = irqflags;
	action->mask = 0;
	action->name = devname;
	action->next = NULL;
	action->dev_id = dev_id;

	retval = tivo_setup_irq(irq, action);

	if (retval)
		kfree(action);
	return retval;
}

void free_irq(unsigned int irq, void *dev_id)
{
	struct irqaction *action, **p;
	unsigned long flags;

	if (DEBUG_IRQS >= 1) {
		printk("free irq %d\n", irq);
	}
	if (irq >= BCM_NR_IRQS) {
		printk(KERN_ERR "Error - trying to free invalid IRQ %d\n", irq);
		return;
	}
	for (p = irq + irq_action; (action = *p) != NULL; p = &action->next) {
		if (action->dev_id != dev_id)
			continue;

		// Found it - now free it
		save_and_cli(flags);
		*p = action->next;
		if (!irq[irq_action]) {
           mask_irq(irq);
        }
		restore_flags(flags);
		kfree(action);
		return;
	}
	printk(KERN_ERR "Trying to free free IRQ%d\n", irq);
}

unsigned long probe_irq_on(void)
{
	// not implemented
	return 0;
}

unsigned int probe_irq_mask(unsigned long val)
{
	// not implemented
	return 0;
}

int probe_irq_off(unsigned long irqs)
{
	// not implemented
	return 0;
}

int (*irq_cannonicalize)(int irq);

static int bcm_irq_cannonicalize(int irq)
{
	return irq;
}

static struct irqaction cascade =
	 { NULL, SA_INTERRUPT, 0, "cascade", NULL, NULL };

void __init init_IRQ(void)
{
	unsigned int status;

	irq_cannonicalize = bcm_irq_cannonicalize;
	set_except_vector(0, tivo_handle_irq);

	status = read_32bit_cp0_register(CP0_STATUS);
	status &= 0xffff00ff;
	write_32bit_cp0_register(CP0_STATUS, status);

	//Interrupts bits are enabled while registering the 
	//interrupt handler. So we disable all interrupts at this point

	// Ack and disable all WIN1 PCI, EXT, & SATA ints
	*BCM_INTR_WIN1_MASK_SET = 0xffffffff;
	barrier();

	// Ack and disable all WIN0 BCM7038 ints
	*BCM_INTR_WIN0_MASK_SET = 0xffffffff;
	barrier();

	*BCM_INTR2_CPU_MASK_SET = 0xffffffff;
	barrier();

	tivo_setup_irq(CPU_IRQ_INT0, &cascade);
	tivo_setup_irq(BCM_IRQ_UPG_CPU_INTR, &cascade);
	tivo_setup_irq(BCM_IRQ_UPG_UART0_CPU_INTR, &cascade);
#ifndef I2C_COMPATABLE_TEMP
    tivo_setup_irq(BCM_IRQ_UPG_BSC_CPU_INTR, &cascade);
#endif
}

inline
static void do_irq0_irq(struct pt_regs *regs)
{
    int bit;
    unsigned int intr_stat;

    intr_stat  = *BCM_INTR_IRQ0_STATUS;
    intr_stat &= *BCM_INTR_IRQ0_ENABLE;

    if (intr_stat) {
        bit = get_leading_one_bit_number(intr_stat);
        do_IRQ( 72 + bit, regs );
        return;
    }

    if (DEBUG_IRQS >= 1)
    {
        printk(KERN_WARNING "spurious BCM int! BCM_INTR_IRQ0_ENABLE=0x%08x  BCM_INTR_IRQ0_STATUS=0x%08x \n",
                            *BCM_INTR_IRQ0_ENABLE, *BCM_INTR_IRQ0_STATUS);
	}
    atomic_inc(&irq_err_count);
}

/*
 *  Handle Broadcom Serial Control ("bsc") interrupt (for I2C)
 */
inline
static void do_irq_bsc(struct pt_regs *regs)
{
    int           irq0_bit;
    unsigned long irq0_intr_stat;

#ifdef I2C_COMPATABLE_TEMP
    // Checking: is there ISR registered for BCM_IRQ_UPG_BSC_CPU_INTR IRQ
	struct irqaction **ppIrqAction;

	ppIrqAction = irq_action + BCM_IRQ_UPG_BSC_CPU_INTR;
    if (*ppIrqAction != NULL) {
        do_IRQ( BCM_IRQ_UPG_BSC_CPU_INTR, regs );
    }
    else { 
#endif 

    irq0_intr_stat  = *BCM_INTR_IRQ0_STATUS;
    irq0_intr_stat &= *BCM_INTR_IRQ0_ENABLE;

    irq0_intr_stat &= ( (1L << (BCM_IRQ_BSC_MASTERA_CPU_INTR - 72)) | 
                        (1L << (BCM_IRQ_BSC_MASTERB_CPU_INTR - 72)) | 
                        (1L << (BCM_IRQ_BSC_MASTERC_CPU_INTR - 72)) | 
                        (1L << (BCM_IRQ_BSC_MASTERD_CPU_INTR - 72))
                      );

    if (irq0_intr_stat) {
        irq0_bit = get_leading_one_bit_number(irq0_intr_stat);
        do_IRQ( irq0_bit + 72, regs );
        return;
    }
#ifdef I2C_COMPATABLE_TEMP
    }
#endif

    if (DEBUG_IRQS >= 1)
    {
        printk(KERN_WARNING "spurious BSC int! BCM_INTR_IRQ0_ENABLE=0x%08x  BCM_INTR_IRQ0_STATUS=0x%08x \n",
                            *BCM_INTR_IRQ0_ENABLE, *BCM_INTR_IRQ0_STATUS);
    }
    atomic_inc(&irq_err_count);
}

inline 
static void do_7038irq(struct pt_regs *regs)
/*
 * This ISR handles the special processing required to handle
 * the (potentially) numerous cascaded handlers associated with 
 * the 7038.  We handle IRQ0 interrupts followed by IRQ1 interrupts.
 * Essentially, we work backwards from handling the SATA interrupt, then
 * the 5 external slot PCI device interrupts, followed by the 3 PCI INTA
 * device interrupts by finally processing the 7038 level 1 device interrupts.
 * Win1 ints followed by Win0. 
 * 
 * Processing interrupts at this level is really just dispatching them
 * to any installed handlers by means of the do_IRQ() function.
 */
{
	int          bit;
	unsigned int intr_stat;

	intr_stat  = *BCM_INTR_WIN0_STATUS;
	intr_stat &= ~(*BCM_INTR_WIN0_MASK_STATUS);

	if (intr_stat )
	{
        bit = get_leading_one_bit_number(intr_stat);
        switch (bit + 8) {
        case BCM_IRQ_UPG_UART0_CPU_INTR:
            do_IRQ( BCM_IRQ_UARTA_CPU_INTR, regs );
            return;
        case BCM_IRQ_UPG_CPU_INTR:
            do_irq0_irq( regs );
            return;
        case BCM_IRQ_UPG_BSC_CPU_INTR:
            do_irq_bsc( regs );
            return;
        default:
            do_IRQ( bit + 8, regs );
            return;
        }
    }

	intr_stat  = *BCM_INTR_WIN1_STATUS;
	intr_stat &= ~(*BCM_INTR_WIN1_MASK_STATUS);

	if (intr_stat )
	{
       bit = get_leading_one_bit_number(intr_stat);
       do_IRQ( 40 + bit, regs );
       return;
    }

	if (DEBUG_IRQS >= 1)
	{
	    printk(KERN_WARNING "spurious BCM int! BCM_INTR_WIN0_STATUS=0x%08x  BCM_INTR_WIN1_STATUS=0x%08x \n",
	       *BCM_INTR_WIN0_STATUS, *BCM_INTR_WIN1_STATUS);
	}
	atomic_inc(&irq_err_count);
}

asmlinkage void do_vector(int irqmask, struct pt_regs *regs)
{
	int irqmask_shifted, bit;

    irqmask_shifted = irqmask >> 8;
    irqmask_shifted &= 0xff;

    if (irqmask_shifted) {
        bit = get_leading_one_bit_number(irqmask_shifted);
        switch (bit) 
        {
        case CPU_IRQ_INT0:
            do_7038irq(regs);
            break;
        default:
            do_IRQ( bit, regs );
            break;
        }
        return;
    }

    if (DEBUG_IRQS >= 4) {
        printk(KERN_WARNING "do_vector got spurious int with mask 0x%08x!\n", irqmask);
    }

    atomic_inc(&irq_err_count);
}

void init_irq_proc (void)
{
}
